"""Custom Exceptions Module."""


class InvalidMessageInputError(ValueError):
    """Exception for invalid inputs in message classes."""

    def __init__(self, msg_type, attr, input_type):
        self.err = 'Invalid input for specified message class: ' + msg_type
        self.err += '\n\t* argument: "{}"'.format(attr)
        self.err += '\n\t* input type must be: {}'.format(input_type)
        super(InvalidMessageInputError, self).__init__(self.err)


class UnsupportedMessageTypeError(TypeError):
    """Exception for declaring unsupported message types."""

    def __init__(self, msg_type, msg_types=None):
        self.err = 'Invalid message type: ' + msg_type
        if msg_types:
            self.err += '\n\t* Supported message types: '
            self.err += str(msg_types)
        super(UnsupportedMessageTypeError, self).__init__(self.err)


class UnknownProfileError(KeyError):
    """Exception for unknown config.json profile names."""

    def __init__(self, profile):
        self.err = 'Unknown Profile name: ' + profile
        super(UnknownProfileError, self).__init__(self.err)



"""messages.slack tests."""

import pytest

import requests

from collections import deque
from unittest.mock import patch

from messages.slack import SlackWebhook
from messages._eventloop import MESSAGELOOP


##############################################################################
# FIXTURES
##############################################################################

@pytest.fixture()
def get_slack(cfg_mock):
    """Return a valid SlackWebhook object."""
    return SlackWebhook(url='test_url', body='message',
                attachments=['url1', 'url2'])


##############################################################################
# TESTS: SlackWebhook.__init__
##############################################################################

def test_slack_init(get_slack):
    """
    GIVEN a need for a SlackWebhook object
    WHEN instantiated
    THEN assert it is properly created
    """
    s = get_slack
    assert s.url == 'test_url'
    assert s.body == 'message'
    assert s.attachments == ['url1', 'url2']
    assert isinstance(s.message, dict)
    assert isinstance(s.sent_messages, deque)


##############################################################################
# TESTS: SlackWebhook.construct_message
##############################################################################

@patch.object(SlackWebhook, 'add_attachments')
def test_slack_construct_message(add_mock, get_slack):
    """
    GIVEN a valid SlackWebhook object
    WHEN construct_message() is called
    THEN assert the message is properly created
    """
    s = get_slack
    s.construct_message()
    assert s.message['text'] == 'message'
    assert add_mock.call_count == 1


@patch.object(SlackWebhook, 'add_attachments')
def test_slack_construct_message_withFromSubj(add_mock, get_slack):
    """
    GIVEN a valid SlackWebhook object
    WHEN construct_message() is called
    THEN assert the message is properly created
    """
    s = get_slack
    s.from_ = 'me'
    s.subject = 'Tst Msg'
    s.construct_message()
    expected = 'From: me\nSubject: Tst Msg\nmessage'
    assert s.message['text'] == expected
    assert add_mock.call_count == 1


##############################################################################
# TESTS: SlackWebhook.add_attachments
##############################################################################

def test_slack_add_attachments_list(get_slack):
    """
    GIVEN a valid SlackWebhook object with self.attach_urls = list
    WHEN add_attachments() is called
    THEN assert the urls are properly attached to the message
    """
    s = get_slack
    s.add_attachments()
    expected = [{'image_url': 'url1', 'author_name': ''},
                {'image_url': 'url2', 'author_name': ''}]
    assert s.message['attachments'] == expected


def test_slack_add_attachments_str(get_slack):
    """
    GIVEN a valid SlackWebhook object with self.attach_urls = str
    WHEN add_attachments() is called
    THEN assert the urls are properly attached to the message
    """
    s = get_slack
    s.attachments = 'url1'
    s.add_attachments()
    expected = [{'image_url': 'url1', 'author_name': ''}]
    assert s.message['attachments'] == expected


def test_slack_add_attachments_with_params(get_slack):
    """
    GIVEN a valid SlackWebhook object with extra attachment params
    WHEN add_attachments() is called
    THEN assert the extra params are properly added
    """
    s = get_slack
    s.attachments = 'url1'
    s.params = {'author_name': 'me', 'text': 'image of me'}
    s.add_attachments()
    expected = [{'image_url': 'url1', 'author_name': 'me',
                'text': 'image of me'}]
    assert s.message['attachments'] == expected


##############################################################################
# TESTS: SlackWebhook.send
##############################################################################

@patch.object(requests, 'post')
@patch.object(SlackWebhook, 'construct_message')
def test_slack_send(con_mock, req_mock, get_slack, capsys):
    """
    GIVEN a valid SlackWebhook object
    WHEN send() is called
    THEN assert the proper send sequence occurs
    """
    s = get_slack
    s.send()
    out, err = capsys.readouterr()
    assert con_mock.call_count == 1
    assert req_mock.call_count == 1
    assert out == 'Message sent...\n'
    assert err == ''
    assert len(s.sent_messages) == 1


##############################################################################
# TESTS: SlackWebhook.send_async
##############################################################################

@patch.object(MESSAGELOOP, 'add_message')
def test_slack_send_async(msgloop_mock, get_slack):
    """
    GIVEN a valid SlackWebhook object
    WHEN send_async() is called
    THEN assert the proper send sequence occurs
    """
    s = get_slack
    s.send_async()
    assert msgloop_mock.call_count == 1



"""Interfaces for message classes."""


from abc import ABCMeta
from abc import abstractmethod

from ._utils import VALIDATOR


class Message(metaclass=ABCMeta):
    """Interface for standard message classes."""

    @abstractmethod
    def send(self):
        """Send message synchronously."""


    @abstractmethod
    def send_async(self):
        """Send message asynchronously."""


    def __setattr__(self, attr, val):
        """Validate attribute inputs after assignment."""
        self.__dict__[attr] = val
        VALIDATOR.validate_input(self, attr)


    def __repr__(self):
        return ('<messages.' + self.__class__.__name__ +
                ' class> at: ' + str(id(self)))


    def __iter__(self):
        return iter(self.__dict__)



"""
Utility Module - globals, classes, and functions useful to all other modules.

1.  Validator
    - Class used to validate user input for each message class.
    - i.e. validate email address inputs or phone number inputs
        using regular expressions.
    - Uses the third-party module, 'validus', for regular expressions.
        https://github.com/shopnilsazal/validus
"""

import functools
import inspect

import validus

from .exceptions import InvalidMessageInputError


class Validator:
    """
    Input validation class.

    Each message classe's 'command' should be a command that returns a bool.
    """

    EMAIL = {
             'from_': {
                       'command': validus.isemail,
                       'type': 'email address',
                       },
             'to': {
                    'command': validus.isemail,
                    'type': 'email address',
                    },
             'cc': {
                    'command': validus.isemail,
                    'type': 'email address or list of email addresses',
                    },
             'bcc': {
                     'command': validus.isemail,
                     'type': 'email address or list of email addresses',
                     },
            }

    TWILIO = {
              'from_': {
                        'command': validus.isphone,
                        'type': 'phone number',
                        },
              'to': {
                     'command': validus.isphone,
                     'type': 'phone number',
                     },
              'media_url': {
                            'command': validus.isurl,
                            'type': 'url',
                    },
            }

    SLACKWEBHOOK = {
                    'webhook_url': {
                                    'command': validus.isurl,
                                    'type': 'url',
                                    },
                    'attach_urls': {
                                    'command': validus.isurl,
                                    'type': 'url or list of urls',
                                    },
                    }

    def __init__(self):
        pass


    def validate_input(self, msg, attr):
        """
        Validate the given input for correct types, using
        regular expressions with the 'validus' package.  Actual validation
        is done via the is_valid() method so the functools.lru_cache decorator
        can be used to memoize/cache inputs.

        Args:
            :msg: object, any instantiated message class
                i.e. e = Email(*args)
            :attr: str, the attribute to validate

        Returns:
            None, but raises InvalidMessageInputError if Validator._is_valid()
                returns False.

        Usage:
            This is used by the __setattr__ override in the
            _interface.Message class, inherited by message classes.
        """
        msgtype = msg.__class__.__name__.upper()
        inputs = getattr(msg, attr)

        if attr in getattr(Validator, msgtype).keys() and inputs is not None:

            if isinstance(inputs, list):
                for i in inputs:
                    if not self._is_valid(msgtype, attr, i):
                        raise InvalidMessageInputError(msg.__class__.__name__,
                            i, getattr(Validator, msgtype)[attr]['type'])
            else:
                if not self._is_valid(msgtype, attr, inputs):
                    raise InvalidMessageInputError(msg.__class__.__name__,
                            inputs, getattr(Validator, msgtype)[attr]['type'])


    @staticmethod
    @functools.lru_cache(maxsize=256)
    def _is_valid(msgtype, attr, val):
        """
        Check each individual input and cache the result via the
        functools.lru_cache decorator to speed up subsequent calls with
        the same input.

        Args:
            :msgtype: str.upper(), type of message in uppercase format
            :attr:, str, the attribute to validate
            :val:, str, the value of that attribute (attr)

        Returns:
            bool, from operation of `command(val)`.  i.e. True if the
                'val' is valid, False otherwise.

        Example inputs:
            Validator._is_valid('EMAIL', 'from_', 'me@here.com')
            returns True
        """
        msgtype = getattr(Validator, msgtype)
        command = msgtype[attr]['command']
        return command(val)


VALIDATOR = Validator()



"""
messages - A package designed to make sending various types of messages easy.
"""

from .api import send
from .email_ import Email
from .slack import SlackWebhook
from .text import Twilio


MESSAGES = ('email', 'slackwebhook', 'twilio')

__version__ = '0.3.0'



"""messages.text tests."""

import pytest

from collections import deque
from unittest.mock import patch

from twilio.rest import Client

import messages.text
from messages.text import Twilio
from messages.text import configure
from messages._eventloop import MESSAGELOOP


##############################################################################
# FIXTURES
##############################################################################

@pytest.fixture()
@patch.object(messages.text, 'configure')
def get_twilio(config_mock):
    """Return a valid Twilio object."""
    t = Twilio(from_='+16198675309', to='+16195551212', acct_sid='test_sid',
        auth_token='test_token', body='test text!',
        attachments='https://imgs.xkcd.com/comics/python.png',
        profile='tester', save=False)
    t.from_ = '+16198675309'
    t.acct_sid = 'test_sid'
    t.auth_token = 'test_token'
    t.profile = 'tester'
    return t


##############################################################################
# TESTS: Twilio.__init__
##############################################################################

def test_twilio_init(get_twilio, cfg_mock):
    """
    GIVEN a need to create an Twilio object
    WHEN the user instantiates a new object with required args
    THEN assert Twilio object is created with given args
    """
    t = get_twilio
    assert t.from_ == '+16198675309'
    assert t.to == '+16195551212'
    assert t.acct_sid == 'test_sid'
    assert t.auth_token == 'test_token'
    assert t.body == 'test text!'
    assert t.attachments == 'https://imgs.xkcd.com/comics/python.png'
    assert t.client is None
    assert isinstance(t.sent_messages, deque)


##############################################################################
# TESTS: Twilio.__str__
##############################################################################

def test_twilio_str(get_twilio,cfg_mock, capsys):
    """
    GIVEN a valid Twilio object
    WHEN the user calls print() on the Twilio object
    THEN assert the correct format prints
    """
    t = get_twilio
    expected = ('Twilio Text Message:'
                '\n\tFrom: +16198675309'
                '\n\tTo: +16195551212'
                '\n\tbody: test text!...'
                '\n\tattachments: https://imgs.xkcd.com/comics/python.png\n')
    print(t)
    out, err = capsys.readouterr()
    assert out == expected
    assert err == ''


##############################################################################
# TESTS: Twilio.get_client
##############################################################################

@patch.object(messages.text, 'Client')
def test_get_client(get_twilio):
    """
    GIVEN a valid Twilio object
    WHEN Twilio.get_client is called
    THEN assert a twilio.rest.Client is returned (a Mock object)
    """
    t = get_twilio
    client = t.get_client()
    assert client is not None


##############################################################################
# TESTS: Twilio.send
##############################################################################

@patch.object(Client, 'messages')
def test_send(messages_mock, get_twilio, cfg_mock, capsys):
    """
    GIVEN a valid Twilio object
    WHEN Twilio.send() is called
    THEN assert the correct functions are called and correct attributes
        updated
    """
    t = get_twilio
    t.send()
    t.sid = 12345
    out, err = capsys.readouterr()
    assert out == 'Message sent...\n'
    assert err == ''
    assert len(t.sent_messages) == 1


##############################################################################
# TESTS: Twilio.send_async
##############################################################################

@patch.object(MESSAGELOOP, 'add_message')
def test_send_async(msg_loop_mock, cfg_mock, get_twilio):
    """
    GIVEN a valid Twilio object
    WHEN Twilio.send_async() is called
    THEN assert it is added to the event loop for async sending
    """
    t = get_twilio
    t.send_async()
    assert msg_loop_mock.call_count == 1



"""
Module designed to make creating and sending emails easy.

1.  Email
    - Uses the Python 3 standard library MIMEMultipart email
      object to construct the email.
"""

import smtplib
import sys
from collections import deque
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication

from .config import configure
from ._eventloop import MESSAGELOOP
from ._interface import Message


SMTP_SERVERS = {
    'gmail.com': ('smtp.gmail.com', 465),
    'yahoo.com': ('smtp.yahoo.com', 465),
    }


class Email(Message):
    """
    Create and send emails using the built-in email package.

    Args:
        :from_: (str) originating email address
            i.e. 'me@here.com'
        :to: (str or list) primary message recipients
             i.e. 'you@there.com' or
                  ['her@there.com', 'him@there.com']
        :server: (str) url of smtp server
            i.e. 'smtp.gmail.com'
        :port: (int) smtp server port
            i.e. 465 or 587
        :password: (str) password for email account
        :cc: (str or list) carbon-copy recipients
        :bcc: (str or list) blind carbon-copy recipients
        :subject: (str) email message subject line
        :body: (str) body text of the message to send
        :attachments: (str or list) files to attach
            i.e. './file1', or
                ['/home/you/file1.txt', '/home/you/file2.pdf']
        :profile: (str) use an account profile specified by name
        :save: (bool) save pertinent values in the messages config.json file,
            such as from_, server, port, password (encrypted keyring) to make
            sending messages faster.

    Attributes:
        :message: (MIMEMultipart) current form of the message to be constructed
        :sent_messages: (deque) all messages sent with current SlackWebHook
            object, acting as a log of messages sent in the current session.

    Usage:
        Create an email object with required Args above.
        Send email with self.send() or self.send_async() methods.

    Note:
        Some email servers may require you to modify security setting, such as
        gmail allowing "less secure apps" to access the account.  Otherwise
        failure may occur when attempting to send.
    """

    def __init__(
        self, from_=None, to=None, server=None, port=None,
        password=None, cc=None, bcc=None, subject='', body='',
        attachments=None, profile=None, save=False
    ):

        config_kwargs = {'from_': from_,
                'server': server or self.get_server(from_)[0],
                'port': port or self.get_server(from_)[1],
                'password': password, 'profile': profile, 'save': save}

        configure(self, params=config_kwargs,
                to_save={'from_', 'server', 'port'}, credentials={'password'})

        self.to, self.cc, self.bcc = to, cc, bcc
        self.subject = subject
        self.body = body
        self.attachments = attachments or []
        self.message = None
        self.sent_messages = deque()


    def __str__(self):
        """print(Email(**args)) method."""
        return('MIMEMultipart Email:'
               '\n\tServer: {}:{}'
               '\n\tFrom: {}'
               '\n\tTo: {}'
               '\n\tCc: {}'
               '\n\tBcc: {}'
               '\n\tSubject: {}'
               '\n\tbody: {}...'
               '\n\tattachments: {}'
               .format(self.server, self.port, self.from_, self.to,
                       self.cc, self.bcc, self.subject, self.body[0:40],
                       self.attachments))


    @staticmethod
    def get_server(address):
        """Return an SMTP servername guess from outgoing email address."""
        if address:
            domain = address.split('@')[1]
            try:
                return SMTP_SERVERS[domain]
            except KeyError:
                return ('smtp.' + domain, 465)
        return (None, None)


    @staticmethod
    def list_to_string(recipient):
        """
        Format the recipient for the MIMEMultipart() email type.
        If the recipient is a list, then it returns the list as a
        comma separated string.
        example: input=['you@here.com', 'her@there.com']
                 output='you@there.com, her@there.com'
        """
        if recipient:
            if isinstance(recipient, list):
                return ', '.join(recipient)
            return recipient


    def generate_email(self):
        """Put the parts of the email together."""
        self.message = MIMEMultipart()
        self.add_header()
        self.add_body()
        self.add_attachments()


    def add_header(self):
        """Add email header info."""
        self.message['From'] = self.from_
        self.message['Subject'] = self.subject


    def add_body(self):
        """Add body content of email."""
        if self.body:
            b = MIMEText('text', 'plain')
            b.set_payload(self.body)
            self.message.attach(b)


    def add_attachments(self):
        """Add required attachments."""
        num_attached = 0
        if self.attachments:
            if isinstance(self.attachments, str):
                self.attachments = [self.attachments]

            for item in self.attachments:
                doc = MIMEApplication(open(item, 'rb').read())
                doc.add_header('Content-Disposition', 'attachment',
                               filename=item)
                self.message.attach(doc)
                num_attached += 1
        return num_attached


    def get_session(self):
        """Start session with email server."""
        if self.port == 465:
            session = self.get_ssl()
        elif self.port == 587:
            session = self.get_tls()
        session.login(self.from_, self.password)
        return session


    def get_ssl(self):
        """Get an SMTP session with SSL."""
        return smtplib.SMTP_SSL(self.server, self.port)


    def get_tls(self):
        """Get an SMTP session with TLS."""
        session = smtplib.SMTP(self.server, self.port)
        session.ehlo()
        session.starttls()
        session.ehlo()
        return session


    def send(self):
        """
        Send the message.
        Append the repr(self) to self.sent_messages as a history.
        """
        self.generate_email()
        session = self.get_session()

        recipients = []
        if self.to:
            self.message['To'] = self.list_to_string(self.to)
            recipients.append(self.to)
        if self.cc:
            self.message['Cc'] = self.list_to_string(self.cc)
            recipients.append(self.cc)
        if self.bcc:
            self.message['Bcc'] = self.list_to_string(self.bcc)
            recipients.append(self.bcc)

        session.sendmail(self.from_, recipients, self.message.as_string())
        session.quit()
        print('Message sent...', file=sys.stdout)
        self.sent_messages.append(repr(self))


    def send_async(self):
        """Send message asynchronously."""
        MESSAGELOOP.add_message(self)
